"""Autogenerated file: DO NOT EDIT!"""
import typing
from dazzler.system import Component, Aspect, UNDEFINED


class Html(Component):
    """
    Html tag wrapper, give any props as `attributes`. Listen to events with
    the readonly event aspect containing the latest event fired.

    """
    _aspect_keys = []
    _children = []
    tag = Aspect(required=True)
    children = Aspect(children=True)
    id = Aspect()
    class_name = Aspect()
    attributes = Aspect()
    events = Aspect()
    event = Aspect()

    def __init__(
            self,
            tag: str,
            children: typing.Optional[typing.Union[str, int, float, Component,typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            id: typing.Optional[str] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            attributes: typing.Optional[typing.Dict] = UNDEFINED,
            events: typing.Optional[typing.List] = UNDEFINED,
            event: typing.Optional[typing.Dict] = UNDEFINED,
            identity: str = None
    ):
        """
        :param tag: Tag name of the component.
        :param children: Children of the html tag.
        :param id: Id of the element in DOM.
        :param class_name: Class of the element.
        :param attributes: Any other html attributes relevant to the html tag
        :param events: Events to subscribe.
        :param event: Last event fired.
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing
from dazzler.system import Component, Aspect, UNDEFINED


class Container(Component):
    """
    Virtual div

    """
    _aspect_keys = []
    _children = []
    children = Aspect(children=True)
    id = Aspect()
    class_name = Aspect()
    style = Aspect()
    title = Aspect()
    draggable = Aspect()
    n_clicks = Aspect(default=0)

    def __init__(
            self,
            children: typing.Optional[typing.Union[str, int, float, Component,typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            id: typing.Optional[str] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            title: typing.Optional[str] = UNDEFINED,
            draggable: typing.Optional[bool] = UNDEFINED,
            n_clicks: typing.Optional[typing.Union[float, int]] = 0,
            identity: str = None
    ):
        """
        :param children:
        :param id:
        :param class_name:
        :param style:
        :param title:
        :param draggable:
        :param n_clicks: (default=0)
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing
from dazzler.system import Component, Aspect, UNDEFINED


class Markdown(Component):
    """
    A react-markdown wrapper for dazzler.

    """
    _aspect_keys = []
    _children = []
    source = Aspect(required=True)
    class_name = Aspect()
    style = Aspect()
    escape_html = Aspect(default=True)
    skip_html = Aspect()
    source_pos = Aspect()
    allowed_types = Aspect()
    disallowed_types = Aspect()
    unwrap_disallowed = Aspect()
    link_target = Aspect()

    def __init__(
            self,
            source: str,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            escape_html: typing.Optional[bool] = True,
            skip_html: typing.Optional[bool] = UNDEFINED,
            source_pos: typing.Optional[bool] = UNDEFINED,
            allowed_types: typing.Optional[typing.List] = UNDEFINED,
            disallowed_types: typing.Optional[typing.List] = UNDEFINED,
            unwrap_disallowed: typing.Optional[bool] = UNDEFINED,
            link_target: typing.Optional[str] = UNDEFINED,
            identity: str = None
    ):
        """
        :param source: Markdown string to render.
        :param class_name: CSS class of the container element.
        :param style: Style object to give to container.
        :param escape_html: Escape html element in the source. (default=True)
        :param skip_html: Skip rendering of html elements.
        :param source_pos: Add data-sourcepos attributes to elements.
        :param allowed_types: Which type of node to allow rendering, default to all.
        :param disallowed_types: Which types of nodes should not be rendered, default
            to none.
        :param unwrap_disallowed: Setting to true will try to extract/unwrap the
            children of disallowed nodes. For instance, if
            disallowing Strong, the default behaviour is to simply
            skip the text within the strong altogether, while the
            behaviour some might want is to simply have the text
            returned without the strong wrapping it.
        :param link_target: Set the default target for link tags. (`_blank` to
            open a new tab)
        """
        Component.__init__(self, locals(), identity)



import os
import json

from .tools import get_package_path

root_path = get_package_path('dazzler')
assets_path = os.path.abspath(os.path.join(root_path, 'assets'))
assets_dist_path = os.path.join(assets_path, 'dist')
assets_dev_path = os.path.join(assets_path, 'dev')
vendors_path = os.path.join(assets_path, 'vendors')


with open(os.path.join(assets_dist_path, 'assets.json')) as _file:
    assets = json.load(_file)['chunks']

with open(os.path.join(assets_dev_path, 'assets.json')) as _file:
    assets_dev = json.load(_file)['chunks']



"""Autogenerated file: DO NOT EDIT!"""
import typing
from dazzler.system import Component, Aspect, UNDEFINED


class Grid(Component):
    """
    Render children in a grid.

    """
    _aspect_keys = []
    _children = []
    children = Aspect(required=True, children=True)
    columns = Aspect(required=True)

    def __init__(
            self,
            children: typing.List[typing.Union[str, int, float, Component,typing.List[typing.Union[str, int, float, Component]]]],  # noqa: E501
            columns: typing.Union[float, int],
            identity: str = None
    ):
        """
        :param children: Children to render in a grid.
        :param columns: Number of columns
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing
from dazzler.system import Component, Aspect, UNDEFINED


class Slider(Component):
    """
    A slider with a caret contained within a min and max value.  Available CSS
    classes (dazzler-core-slider scoped): - slider-area: The outer div of the
    slider. - slider-handle: The drag handle - slider-spacer: The space that
    is before the handle.

    """
    _aspect_keys = []
    _children = []
    minimum = Aspect(required=True)
    maximum = Aspect(required=True)
    value = Aspect(default=0)
    round = Aspect()
    debounce = Aspect(default=50)
    class_name = Aspect()
    style = Aspect()

    def __init__(
            self,
            minimum: typing.Union[float, int],
            maximum: typing.Union[float, int],
            value: typing.Optional[typing.Union[float, int]] = 0,
            round: typing.Optional[typing.Any] = UNDEFINED,
            debounce: typing.Optional[typing.Union[float, int]] = 50,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            identity: str = None
    ):
        """
        :param minimum: Minimum (leftmost) value of the slider.
        :param maximum: Maximum (rightmost) value of the slider.
        :param value: Current value (default=0)
        :param round: Round the value (Possible values: 'ceil', 'floor')
        :param debounce: Time in milliseconds to wait before updating the
            value. (default=50)
        :param class_name: CSS classes to use. (Scope: dazzler-core-slider)
        :param style: Style object of the root div.
        """
        Component.__init__(self, locals(), identity)



import argparse
import functools
import importlib
import itertools
import shutil
import sys
import asyncio
import os
import typing

from concurrent.futures import ThreadPoolExecutor

from aiohttp import web
import precept

from .system import Package, generate_components, Requirement, Page
from .tools import get_package_path

from ._config import DazzlerConfig
from ._server import Server
from ._version import __version__
from .errors import PageConflictError, ServerStartedError
from ._assets import assets_path
# noinspection PyProtectedMember
from .system._requirements import _internal_data_dir


class Dazzler(precept.Precept):
    """
    Dazzler web application & cli.

    :Commands:

        - ``dazzler --root application``, start an application.
        - ``dazzler generate components_metadata.json output_dir``
        - ``dazzler copy-requirements -p dazzler_core -p dazzler_extra``
    """
    config: DazzlerConfig
    config_class = DazzlerConfig
    version = __version__
    pages: typing.Dict[str, Page]
    global_arguments = [
        precept.Argument(
            '-a', '--application',
            help='Path to the application',
            type=str
        ),
    ]
    server: Server

    def __init__(self, module_name):
        self.root_path = get_package_path(module_name)
        super().__init__(
            config_file=[
                'dazzler.toml',
                os.path.join(self.root_path, 'dazzler.toml')
            ],
            add_dump_config_command=True,
            executor=ThreadPoolExecutor(),
            print_version=False,
        )
        self.requirements: typing.List[Requirement] = []
        self.server = Server(self, loop=self.loop)
        self.pages = {}
        self.stop_event = asyncio.Event()
        self._prepared = False
        self._started = False

    def add_page(self, *pages: Page):
        if self._started:
            raise ServerStartedError(
                f'Cannot add pages after the server started:'
                f' {",".join(str(x) for x in pages)}'
            )
        for page in pages:
            if page.name in self.pages:
                raise PageConflictError(f'Duplicate page name: {page}')
            for x in self.pages.values():
                if x.url == page.url:
                    raise PageConflictError(f'Duplicate page url: {page}')
            self.pages[page.name] = page

    async def stop(self):
        await self.server.site.stop()

    # pylint: disable=arguments-differ
    async def main(
            self,
            application=None,
            blocking=True,
            debug=False,
            **kwargs
    ):
        if application:
            # When running from the command line need to insert the path.
            # Otherwise it will never find it.
            sys.path.insert(0, '.')
            mod_path, app_name = application.split(':')
            module = importlib.import_module(mod_path)
            app: Dazzler = getattr(module, app_name)
            # Set cli for arguments
            app.cli = self.cli
        else:
            app = self

        await app.setup_server(debug=debug or app.config.debug)

        await app.server.start(
            app.config.host, app.config.port,
        )

        # Test needs it to run without loop
        # Otherwise the server closes when the event loop ends.
        if blocking:
            await self._loop()

    async def _loop(self):
        while not self.stop_event.is_set():
            # Just run so the site doesn't close.
            # Maybe add something later here.
            await asyncio.sleep(100)

    async def setup_server(self, debug=False):
        # Gather global requirements from configs.
        for external in itertools.chain(
                self.config.requirements.external_scripts,
                self.config.requirements.external_styles
        ):
            self.requirements.append(Requirement(external=external))

        for internal in itertools.chain(
                self.config.requirements.internal_scripts,
                self.config.requirements.internal_styles
        ):
            self.requirements.append(Requirement(internal=internal))

        # Copy all requirements to make sure all is latest.
        await self.copy_requirements()

        # Remove the renderer from package_list once copied.
        # Served by default by the index.
        if 'dazzler_renderer' in Package.package_registry:
            Package.package_registry.pop('dazzler_renderer')

        # Add package defined routes
        routes = [x.routes for x in Package.package_registry.values()]

        # Add page specifics routes
        for page in self.pages.values():
            # Pages are not prefixed
            # noinspection PyTypeChecker
            routes += [[
                web.get(
                    page.url,
                    functools.partial(self.server.route_page, page=page),
                    name=page.name,
                ),
                web.post(
                    page.url,
                    functools.partial(self.server.route_page_json, page=page),
                    name=f'{page.name}-api'
                )
            ]]
            routes += [page.routes]

        self.server.setup_routes(
            list(itertools.chain(*routes)),
            debug=debug,
        )

        self._prepared = True
        self.server.app.on_startup.append(self._on_startup)
        self.server.app.on_shutdown.append(self._on_shutdown)
        await self.events.dispatch('dazzler_setup', application=self)

    async def application(self):
        """Call for wsgi application"""
        if not self._prepared:
            await self.setup_server()
        return self.server.app

    # Commands

    @precept.Command(
        precept.Argument(
            'metadata',
            type=argparse.FileType('r'),
            default=sys.stdin,
            help='react-docgen output'
        ),
        precept.Argument(
            'output_dir',
            help='Output the components in this folder.'
        ),
        description='Generate dazzler components from react-docgen output'
    )
    async def generate(self, metadata, output_dir):
        os.makedirs(output_dir, exist_ok=True)
        await generate_components(metadata, output_dir, self.executor)

    @precept.Command(
        precept.Argument('-p', '--packages', nargs='+'),
        description='Copy the application requirements.'
    )
    async def copy_requirements(self, packages=tuple()):
        self.logger.debug('Copying requirements.')
        for package in packages:
            self.logger.debug(f'Importing package: {package}')
            importlib.import_module(package)

        futures = []

        requirements = [
            package.requirements for package
            in Package.package_registry.values()
        ]
        requirements += [self.requirements]
        requirements += [
            page.requirements for page in self.pages.values()
        ]

        for requirement in itertools.chain(*requirements):
            if not requirement.internal:
                continue
            self.logger.debug(
                f'Copying {requirement.internal} '
                f'to {requirement.internal_static}'
            )
            futures.append(self.executor.execute(
                shutil.copy,
                requirement.internal,
                requirement.internal_static
            ))
            if requirement.dev:
                self.logger.debug(
                    f'Copying {requirement.dev} '
                    f'to {requirement.dev_static}'
                )
                futures.append(
                    self.executor.execute(
                        shutil.copy,
                        requirement.dev,
                        requirement.dev_static
                    )
                )

        # Copy the mount script.
        futures.append(
            self.executor.execute(
                shutil.copy,
                os.path.join(assets_path, 'index.js'),
                os.path.join(_internal_data_dir)
            )
        )
        await asyncio.gather(*futures)

    async def _on_startup(self, _):
        await self.events.dispatch('dazzler_start', application=self)

    async def _on_shutdown(self, _):
        await self.events.dispatch('dazzler_stop', application=self)


def cli():
    dazzler = Dazzler('__main__')
    dazzler.start()



"""Autogenerated file: DO NOT EDIT!"""
import typing
from dazzler.system import Component, Aspect, UNDEFINED


class DatePicker(Component):
    """

    """
    _aspect_keys = []
    _children = []
    date = Aspect()
    focused = Aspect()
    id = Aspect()
    placeholder = Aspect()
    aria_label = Aspect()
    disabled = Aspect()
    required = Aspect()
    read_only = Aspect()
    screen_reader_input_message = Aspect()
    show_clear_date = Aspect()
    custom_close_icon = Aspect(children=True)
    show_default_input_icon = Aspect()
    custom_input_icon = Aspect(children=True)
    input_icon_position = Aspect()
    no_border = Aspect()
    block = Aspect()
    small = Aspect()
    regular = Aspect()
    orientation = Aspect()
    anchor_direction = Aspect()
    open_direction = Aspect()
    horizontal_margin = Aspect()
    with_portal = Aspect()
    with_full_screen_portal = Aspect()
    append_to_body = Aspect()
    disable_scroll = Aspect()
    first_day_of_week = Aspect()
    number_of_months = Aspect()
    keep_open_on_date_select = Aspect()
    reopen_picker_on_clear_date = Aspect()
    hide_keyboard_shortcuts_panel = Aspect()
    day_size = Aspect()
    is_RTL = Aspect()
    transition_duration = Aspect()
    display_format = Aspect()
    month_format = Aspect()
    week_day_format = Aspect()
    phrases = Aspect()
    day_aria_label_format = Aspect()

    def __init__(
            self,
            date: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            focused: typing.Optional[bool] = UNDEFINED,
            id: typing.Optional[str] = UNDEFINED,
            placeholder: typing.Optional[str] = UNDEFINED,
            aria_label: typing.Optional[str] = UNDEFINED,
            disabled: typing.Optional[bool] = UNDEFINED,
            required: typing.Optional[bool] = UNDEFINED,
            read_only: typing.Optional[bool] = UNDEFINED,
            screen_reader_input_message: typing.Optional[str] = UNDEFINED,
            show_clear_date: typing.Optional[bool] = UNDEFINED,
            custom_close_icon: typing.Optional[typing.Union[str, int, float, Component,typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            show_default_input_icon: typing.Optional[bool] = UNDEFINED,
            custom_input_icon: typing.Optional[typing.Union[str, int, float, Component,typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            input_icon_position: typing.Optional[typing.Any] = UNDEFINED,
            no_border: typing.Optional[bool] = UNDEFINED,
            block: typing.Optional[bool] = UNDEFINED,
            small: typing.Optional[bool] = UNDEFINED,
            regular: typing.Optional[bool] = UNDEFINED,
            orientation: typing.Optional[typing.Any] = UNDEFINED,
            anchor_direction: typing.Optional[typing.Any] = UNDEFINED,
            open_direction: typing.Optional[typing.Any] = UNDEFINED,
            horizontal_margin: typing.Optional[typing.Union[float, int]] = UNDEFINED,  # noqa: E501
            with_portal: typing.Optional[bool] = UNDEFINED,
            with_full_screen_portal: typing.Optional[bool] = UNDEFINED,
            append_to_body: typing.Optional[bool] = UNDEFINED,
            disable_scroll: typing.Optional[bool] = UNDEFINED,
            first_day_of_week: typing.Optional[typing.Any] = UNDEFINED,
            number_of_months: typing.Optional[typing.Union[float, int]] = UNDEFINED,  # noqa: E501
            keep_open_on_date_select: typing.Optional[bool] = UNDEFINED,
            reopen_picker_on_clear_date: typing.Optional[bool] = UNDEFINED,
            hide_keyboard_shortcuts_panel: typing.Optional[bool] = UNDEFINED,
            day_size: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            is_RTL: typing.Optional[bool] = UNDEFINED,
            transition_duration: typing.Optional[typing.Union[float, int]] = UNDEFINED,  # noqa: E501
            display_format: typing.Optional[str] = UNDEFINED,
            month_format: typing.Optional[str] = UNDEFINED,
            week_day_format: typing.Optional[str] = UNDEFINED,
            phrases: typing.Optional[typing.Dict] = UNDEFINED,
            day_aria_label_format: typing.Optional[str] = UNDEFINED,
            identity: str = None
    ):
        """
        :param date:
        :param focused:
        :param id:
        :param placeholder:
        :param aria_label:
        :param disabled:
        :param required:
        :param read_only:
        :param screen_reader_input_message:
        :param show_clear_date:
        :param custom_close_icon:
        :param show_default_input_icon:
        :param custom_input_icon:
        :param input_icon_position: (Possible values: 'before', 'after')
        :param no_border:
        :param block:
        :param small:
        :param regular:
        :param orientation: (Possible values: 'horizontal', 'vertical')
        :param anchor_direction: (Possible values: 'left', 'right')
        :param open_direction: (Possible values: 'down', 'up')
        :param horizontal_margin:
        :param with_portal:
        :param with_full_screen_portal:
        :param append_to_body:
        :param disable_scroll:
        :param first_day_of_week: (Possible values: 0, 1, 2, 3, 4, 5, 6)
        :param number_of_months:
        :param keep_open_on_date_select:
        :param reopen_picker_on_clear_date:
        :param hide_keyboard_shortcuts_panel:
        :param day_size:
        :param is_RTL:
        :param transition_duration:
        :param display_format:
        :param month_format:
        :param week_day_format:
        :param phrases:
        :param day_aria_label_format:
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing
from dazzler.system import Component, Aspect, UNDEFINED


class Meter(Component):
    """
    A meter is like a progress bar with more options. Unsupported in IE.

    """
    _aspect_keys = []
    _children = []
    value = Aspect(default=0)
    minimum = Aspect(default=0)
    maximum = Aspect(default=1)
    low = Aspect()
    high = Aspect()
    optimum = Aspect()
    id = Aspect()
    class_name = Aspect()
    style = Aspect()
    click = Aspect()

    def __init__(
            self,
            value: typing.Optional[typing.Union[float, int]] = 0,
            minimum: typing.Optional[typing.Union[float, int]] = 0,
            maximum: typing.Optional[typing.Union[float, int]] = 1,
            low: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            high: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            optimum: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            id: typing.Optional[str] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            click: typing.Optional[typing.Dict[str, typing.Union[typing.Union[float, int]]]] = UNDEFINED,  # noqa: E501
            identity: str = None
    ):
        """
        :param value: The current numeric value. This must be between the
            minimum and maximum values (min attribute and max
            attribute) if they are specified. If unspecified or
            malformed, the value is 0. If specified, but not
            within the range given by the min attribute and max
            attribute, the value is equal to the nearest end of
            the range. (default=0)
        :param minimum: The lower numeric bound of the measured range. This
            must be less than the maximum value (max attribute),
            if specified. If unspecified, the minimum value is 0
            (default=0)
        :param maximum: The upper numeric bound of the measured range. This
            must be greater than the minimum value (min
            attribute), if specified. If unspecified, the maximum
            value is 1. (default=1)
        :param low: The upper numeric bound of the low end of the measured
            range. This must be greater than the minimum value (min
            attribute), and it also must be less than the high
            value and maximum value (high attribute and max
            attribute, respectively), if any are specified. If
            unspecified, or if less than the minimum value, the low
            value is equal to the minimum value.
        :param high: The lower numeric bound of the high end of the
            measured range. This must be less than the maximum
            value (max attribute), and it also must be greater than
            the low value and minimum value (low attribute and min
            attribute, respectively), if any are specified. If
            unspecified, or if greater than the maximum value, the
            high value is equal to the maximum value.
        :param optimum: This attribute indicates the optimal numeric value. It
            must be within the range (as defined by the min
            attribute and max attribute). When used with the low
            attribute and high attribute, it gives an indication
            where along the range is considered preferable. For
            example, if it is between the min attribute and the low
            attribute, then the lower range is considered
            preferred.
        :param id: Id for this component, otherwise the identity is used.
        :param class_name: CSS class.
        :param style: Style of the component.
        :param click: Readonly click event with the value.
        """
        Component.__init__(self, locals(), identity)



from precept import Config, ConfigProperty, ConfigFormat, Nestable


class DazzlerConfig(Config):
    """Dazzler configuration"""

    app_title = ConfigProperty(
        default='Dazzler',
        comment='Name of the title of the index.'
    )

    host = ConfigProperty(
        default='127.0.0.1',
        comment='Host address',
        auto_global=True,
    )

    port = ConfigProperty(
        default=8150,
        comment='Port of the server',
        auto_global=True,
    )

    debug = ConfigProperty(
        default=False,
        config_type=bool,
        auto_global=True,
    )

    route_prefix = ConfigProperty(
        default='',
        comment='Route prefix for all dazzler related endpoints.',
    )

    static_folder = ConfigProperty(
        default='static',
        comment='Path relative to project folder where files will be served.',
    )

    static_prefix = ConfigProperty(
        default='/static',
        comment='Prefix for the static route'
    )

    class Requirements(Nestable):
        prefer_external = ConfigProperty(
            default=False,
            config_type=bool,
            comment='Prefer serving external requirements when available'
        )
        external_scripts = ConfigProperty(
            default=[],
            config_type=list,
            comment='List of urls to include as script requirement.',
        )
        internal_scripts = ConfigProperty(
            default=[],
            config_type=list,
            comment='List of files to include as script requirement.',
        )
        external_styles = ConfigProperty(
            default=[],
            config_type=list,
            comment='List of urls to include as style requirement.'
        )
        internal_styles = ConfigProperty(
            default=[],
            config_type=list,
            comment='List of files to include as script requirement.',
        )

    requirements: Requirements

    class Renderer(Nestable):
        retries = ConfigProperty(
            default=20,
            config_type=int,
            comment='Number of times it will try to reconnect '
                    'when the websocket connection is lost.'
        )
        ping = ConfigProperty(
            default=False,
            config_type=bool,
            comment='Enable to send a ping every interval to keep the '
                    'websocket connected if it didn\'t send data after '
                    'a delay. Some hosts providers will automatically closes'
                    'idling connection after a while.'
        )
        ping_interval = ConfigProperty(
            default=25.0,
            config_type=float,
            comment='Interval at which to send ping data.'
        )

    renderer: Renderer

    def __init__(self):
        super().__init__(root_name='dazzler', config_format=ConfigFormat.TOML)



"""
Dazzler systems (API)

- Requirements are JS/CSS resources to include on rendering.
- Packages hold components info and it's requirements.
- Component Aspects are shared between backend and frontend with bindings.
- Generate components with
``dazzler generate metadata.json output_dir``
- Page holds meta data for page rendering and requirements, routes, layout and
bindings used on a single page of the application.
"""
from ._component import Component, Aspect  # noqa: F401
from ._binding import *  # noqa: F401, F403
from ._package import Package  # noqa: F401
from ._requirements import (  # noqa: F401
    Requirement,
    RequirementWarning,
    assets_to_requirements,
    collect_requirements,
)
from ._generator import generate_components  # noqa: F401
from ._undefined import UNDEFINED, Undefined  # noqa: F401
from ._page import Page  # noqa: F401



import asyncio
import os
import pkgutil
import sys
from ssl import SSLContext
from typing import Optional
import weakref

import aiohttp

from aiohttp import web, WSCloseCode

from .system import Page, UNDEFINED

# noinspection PyProtectedMember
from .system._requirements import _internal_data_dir
from .tools import replace_all, format_tag
from ._renderer import package as renderer


class Server:
    """
    Dazzler server

    :Routes:

        - ``/dazzler/update``
        - ``/dazzler/requirements/``

    """
    runner: web.AppRunner
    loop: asyncio.AbstractEventLoop
    app: web.Application
    site: web.TCPSite

    def __init__(self, dazzler, loop=None, app: web.Application = None):
        """
        :param dazzler: Dazzler instance.
        :type dazzler: dazzler.Dazzler
        :param loop: Asyncio event loop.
        :param app: An existing aiohttp web application to use.
        """
        self.dazzler = dazzler
        self.loop = loop or asyncio.get_event_loop()
        self.app = app or web.Application()
        self.index = pkgutil.get_data(
            'dazzler', os.path.join('assets', 'index.html')
        ).decode()
        self.logger = self.dazzler.logger
        self.websockets = weakref.WeakSet()
        self.debug = False

    def setup_routes(self, routes: list = None, debug: bool = False):
        """
        Setup routes for dazzler.

        :param routes: Routes to add the server.
        :param debug: Set debug mode.
        :return:
        """
        routes = routes or []
        prefix = self.dazzler.config.route_prefix
        self.debug = debug

        # Dazzler api.
        self.app.add_routes([
            web.get(f'{prefix}/dazzler/update', self.route_update),
            web.get(f'{prefix}/dazzler/link', self.route_get_page)
        ] + routes)

        # User static directory.
        static_dir = os.path.join(
            self.dazzler.root_path, self.dazzler.config.static_folder
        )
        if os.path.exists(static_dir):
            self.app.router.add_static(
                f'{prefix}{self.dazzler.config.static_prefix}',
                static_dir
            )

        # Main internal static directory where requirements are copied.
        self.app.router.add_static(
            f'{prefix}/dazzler/requirements/static',
            _internal_data_dir,
        )

    async def route_update(self, request: web.Request):
        """
        WebSocket route for aspect updating.

        :param request: The incoming request.
        :return:
        """
        ws = web.WebSocketResponse()

        await ws.prepare(request)

        self.websockets.add(ws)

        request_queue = asyncio.Queue()
        pendings = []
        aspect_requests = {}
        storage_requests = {}
        done = asyncio.Event()

        async def request_loop():
            while not done.is_set():
                req = await request_queue.get()
                kind = req['kind']
                if kind == 'get-aspect':
                    aspect_requests[req['request_id']] = req.pop('queue')
                elif kind == 'get-storage':
                    storage_requests[req['request_id']] = req.pop('queue')
                await ws.send_json(req)

        def done_callback(done: asyncio.Task):
            pendings.remove(done)
            exception = done.exception()
            if exception and not done.cancelled():
                done.print_stack(file=sys.stderr)
                self.logger.error(exception)

        async def handler():
            async for msg in ws:  # type: aiohttp.WSMessage
                if msg.type == aiohttp.WSMsgType.TEXT:
                    # noinspection PyNoneFunctionAssignment
                    data = msg.json()
                    kind = data.get('kind')

                    if kind == 'binding':
                        page: Page = self.dazzler.pages[data['page']]
                        binding = page.get_binding(data['key'])

                        task = self.loop.create_task(
                            binding(request, data, ws, request_queue)
                        )
                        pendings.append(task)
                        task.add_done_callback(done_callback)

                    elif kind == 'get-aspect':
                        request_id = data['request_id']
                        value = data.get('value', UNDEFINED)
                        req = aspect_requests.pop(request_id)

                        if value is UNDEFINED:
                            await req.put((UNDEFINED, data.get('error')))
                        else:
                            await req.put((value, UNDEFINED))

                    elif kind == 'get-storage':
                        request_id = data['request_id']
                        value = data.get('value', UNDEFINED)
                        queue = storage_requests.pop(request_id)
                        await queue.put((value, UNDEFINED))
                else:
                    self.logger.debug(f'No handler for msg type: {msg.type}')

            done.set()

        async def pong():
            while not done.is_set():
                await asyncio.sleep(self.dazzler.config.renderer.ping_interval)
                await ws.send_json({'kind': 'ping'})

        try:
            ops = [handler(), request_loop()]
            if self.dazzler.config.renderer.ping:
                ops.append(pong())
            await asyncio.gather(*ops)
        finally:
            for pending in pendings:
                pending.cancel()
            self.websockets.discard(ws)

        return ws

    async def route_page(self, _: web.Request, page: Page = None):
        """
        Index route for a page.

        :param _:
        :param page: The page to serve.
        :return:
        """
        script = {
            'src': '/dazzler/requirements/static/index.js',
            'data-retries': str(self.dazzler.config.renderer.retries),
            'id': 'dazzler-script'
        }
        if self.dazzler.config.renderer.ping:
            script.update({
                'data-ping': 'true',
                'data-ping-interval':
                    str(self.dazzler.config.renderer.ping_interval)
            })

        index = replace_all(
            self.index,
            page_title=page.title,
            renderer_scripts='\n'.join(
                x.tag(
                    dev=self.debug,
                    external=self.dazzler.config.requirements.prefer_external
                )
                for x in renderer.requirements if x.kind == 'js'
            ),
            css='\n'.join(
                x.tag(
                    dev=self.debug,
                    external=self.dazzler.config.requirements.prefer_external
                )
                for x in renderer.requirements if x.kind == 'css'
            ),
            dazzler_script=format_tag('script', script),
            meta='\n'.join(format_tag('meta', x) for x in page.meta_tags),
            favicon=format_tag(
                'link', {
                    'rel': 'icon',
                    'href': page.favicon,
                    'type': 'image/x-icon'
                }
            ) if page.favicon else '',
            header=page.header,
            footer=page.footer,
            lang=page.lang,
        )
        return web.Response(body=index, content_type='text/html')

    # noinspection PyMethodMayBeStatic
    async def route_page_json(self, request: web.Request, page: Page = None):
        """
        Serve the bindings and layout associated with page.

        :param request:
        :param page:
        :return:
        """
        prepared = await page.prepare(
            request,
            self.debug,
            external=self.dazzler.config.requirements.prefer_external
        )
        # Add global requirements first.
        prepared['requirements'] = [
            x.prepare(
                dev=self.debug,
                external=self.dazzler.config.requirements.prefer_external
            ) for x in self.dazzler.requirements
        ] + prepared['requirements']
        return web.json_response(prepared)

    async def route_get_page(self, request: web.Request):
        page = request.query.get('page')
        if page in self.dazzler.pages:
            raise web.HTTPFound(location=request.app.router[page].url_for())
        raise web.HTTPNotFound()

    async def start(
            self, host: str, port: int,
            shutdown_timeout: float = 60.0,
            ssl_context: Optional[SSLContext] = None,
            backlog: int = 128,
            reuse_address: Optional[bool] = None,
            reuse_port: Optional[bool] = None,
    ):
        """
        Start the server

        :param host:
        :param port:
        :param shutdown_timeout:
        :param ssl_context:
        :param backlog:
        :param reuse_address:
        :param reuse_port:
        :return:
        """
        self.app.on_shutdown.append(self._on_shutdown)
        self.logger.debug('Starting server')
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        self.site = web.TCPSite(
            self.runner, host, port,
            shutdown_timeout=shutdown_timeout,
            ssl_context=ssl_context, backlog=backlog,
            reuse_address=reuse_address, reuse_port=reuse_port
        )
        await self.site.start()
        self.logger.info(f'Started server http://{host}:{port}/')

    async def _on_shutdown(self, _):
        # Close all websocket's and stop the main loop
        self.dazzler.stop_event.set()
        for ws in set(self.websockets):
            await ws.close(code=WSCloseCode.GOING_AWAY,
                           message='Server shutdown')



"""
Component packages of dazzler.

- ``core``: Lightweight interactive components.
- ``extra``: Bulkier components with more functionality (markdown, syntax, etc)
- ``analytics``: Graphs and viz.
- ``media``: Video, audio,
"""



"""Autogenerated file: DO NOT EDIT!"""
import typing
from dazzler.system import Component, Aspect, UNDEFINED


class Markdown(Component):
    """

    """
    _aspect_keys = []
    _children = []

    def __init__(
            self,
            identity: str = None
    ):
        """
        
        """
        Component.__init__(self, locals(), identity)
